/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/extensions/annotation_ext.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/extension_element_ext.dart';
import 'package:welltested/src/extensions/test_request_extension.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../src/executors/ast_elm_executor.dart';

class UnitGenerator {
  UnitGenerator(
      {required this.dartServerRepository, required this.cacheManager});
  final DartServerRepository dartServerRepository;
  final CacheManager cacheManager;
  late final GenerationRepository generationRepository;
  late final WriterRepository writerRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final AstElementExecutor astElementExecutor;
  late final UnitTestExecutor unitTestExecutor;
  late final WriterExecutor writerExecutor;

  void init() {
    generationRepository = GenerationRepository();
    writerRepository = WriterRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    astElementExecutor = AstElementExecutor();
    unitTestExecutor = UnitTestExecutor(generationRepository);
    writerExecutor =
        WriterExecutor(writerRepository, dartServerRepository, fixerExecutor);
  }

  Future<void> generateForClass(
      {required ClassElement classElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ClassDeclaration classAstNode,
      required PubspecData pubspecData,
      required File classFile,
      required String? requestedFunction,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final methodAstNodes = await classElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      if (requestedFunction != null && method.key != requestedFunction)
        continue;
      await _generateTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          classAstNode: classAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: classFile,
          testRunInfo: testRunInfo,
          className: classElement.displayName,
          methodAstNodes: methodAstNodes);
    }
  }

  Future<void> generateForExtension(
      {required ExtensionElement extElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ExtensionDeclaration extAstNode,
      required PubspecData pubspecData,
      required File extFile,
      required String? requestedFunction,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(extFile.path)}';

    final methodAstNodes = await extElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      if (requestedFunction != null && method.key != requestedFunction)
        continue;
      await _generateTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          extAstNode: extAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: extFile,
          testRunInfo: testRunInfo,
          className: extElement.displayName,
          methodAstNodes: methodAstNodes);
    }
  }

  Future<void> generateForGlobalMethod(
      {required FunctionElement methodElement,
      required FunctionDeclaration methodAstNode,
      required CompilationUnit compilationUnit,
      required ElementAnnotation annotation,
      required PubspecData pubspecData,
      required File methodFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    await _generateTestsForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
        testRunInfo: testRunInfo);
  }

  Future<void> _generateTestsForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    required File subjectFile,
    required TestRunInfo testRunInfo,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      wtLog.updateSpinnerMessage(
          "Generating Test for ${testRequestData.functionName}");
      final file =
          PathUtils.correspondingPermanentTestFile(method.key, subjectFile);

      final existStatus = await cacheManager.testExistStatus(
          testRequestData.codeInScrutity(excludeAnnotations: false),
          subjectFile.path,
          testRequestData.functionName,
          className: testRequestData.parentEntityName);
      bool update = false;
      late TestGenerationStatus testGenerationStatus;
      switch (existStatus) {
        case TestExistStatus.generationExists:
          testGenerationStatus = TestGenerationStatus.omit;
          return;
        case TestExistStatus.onlyRequestExists:
          testRequestData.forceGenerate = true;
          testGenerationStatus = TestGenerationStatus.forceUpdate;
          break;
        case TestExistStatus.outdatedCodeExists:
          update = true;
          testGenerationStatus = TestGenerationStatus.update;
          break;
        case TestExistStatus.doesNotExist:
          testGenerationStatus = TestGenerationStatus.create;
          break;
      }

      final unitTests = await unitTestExecutor.buildUnitTests(testRequestData);
      String? completion = unitTests['fixed_completion'];
      late File testFile;
      if (completion != null) {
        testFile = await writerExecutor.outputGeneration(file.path, completion);
      } else {
        completion = unitTests['model_completion'];
        if (completion == null) {
          wtLog.warning(
              "⛔️ Sorry, we can't process ${testRequestData.functionName} at the moment. Please try again Later.");
          return;
        }
        testFile = await writerExecutor.processGeneration(
            file.path, completion, method.key);
      }
      await cacheManager.cacheTest(
          code: testRequestData.codeInScrutity(excludeAnnotations: false),
          codeFilePath: subjectFile.path,
          test: completion,
          testFilePath: testFile.path,
          className: testRequestData.parentEntityName,
          functionName: testRequestData.functionName);
      testRunInfo.incrementFromAction(testGenerationStatus);
      wtLog.log(
          '✔️ Tests ${update ? 'updated' : 'generated'} for ${method.key}');
    } catch (e, stackTrace) {
      testRunInfo.incrementFromAction(TestGenerationStatus.failed);
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error generating unit tests for method: ${method.key}");
      wtLog.log(e.toString());
    }
  }
}
